from typing import TypeVar

from rainbowadn.chain.states.activestateprotocol import ActiveStateProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.wrisbt.wrisbtindex import WrisbtIndex
from rainbowadn.wrisbt.wrisbtroot import WrisbtRoot

__all__ = ('WrisbtProtocol',)

TargetType = TypeVar('TargetType')


class WrisbtProtocol(ActiveStateProtocol[TargetType, WrisbtIndex]):
    def __init__(self, resolver: HashResolver, keymin: int):
        assert isinstance(resolver, HashResolver)
        assert isinstance(keymin, int)
        super().__init__(resolver)
        assert keymin >= 2
        self.keymin = keymin

    def _initial_state(self) -> HashPoint[WrisbtIndex]:
        return HashPoint.of(
            WrisbtIndex(
                HashPoint.of(WrisbtRoot.empty(self.keymin)),
                HashPoint.of(WrisbtRoot.empty(self.keymin)),
                self.keymin
            )
        )

    def _derive(
            self,
            previous: HashPoint[WrisbtIndex],
            header: HashPoint[TargetType]
    ) -> HashPoint[WrisbtIndex]:
        assert isinstance(previous, HashPoint)
        assert isinstance(header, HashPoint)
        index = self.resolver.resolve(previous)
        assert isinstance(index, WrisbtIndex)

        empty: WrisbtRoot = WrisbtRoot.empty(self.keymin)
        assert isinstance(empty, WrisbtRoot)
        total: WrisbtRoot = self.resolver.resolve(index.total)
        assert isinstance(total, WrisbtRoot)

        return HashPoint.of(
            WrisbtIndex(
                HashPoint.of(total.index(self.resolver, header, empty)),
                HashPoint.of(empty.index(self.resolver, header, total)),
                index.keymin
            )
        )
